import { TelemetryLevel, TelemetryHandler } from './telemetry';
/** The name of the vacancy attribute. */
export declare const VACANCY_ATTRIBUTE = "data-vacancy";
/**
 * A promise that resolves when all vacancies matching
 * a given pattern have disappeared from the DOM.
 */
export declare type ExitPromise = Promise<void>;
/**
 * Key/value pairs representing a match against a vacancy.
 * For example, the vacancy `"users/123"` matches the pattern `"users/:id"`.
 * The corresponding match object would look like this: `{ id: '123' }`.
 */
export interface PatternMatch {
    [key: string]: string;
}
/**
 * Options used when creating a Motel instance.
 */
export interface MotelOptions {
    /**
     * Higher levels generate less telemetry.
     * For example, use `"warn"` in prod, `"debug"` in dev.
     * The default is `"warn"`.
     */
    telemetryLevel?: TelemetryLevel;
    /**
     * Supplying this overrides the default behavior of logging to console.
     */
    telemetryHandler?: TelemetryHandler;
}
/**
 * Callback that executes when a vacancy matches a string pattern.
 *
 * @typeparam A The type of object that you'll dispatch to your app.
 * @param matchObject The result of matching the vacancy with the string pattern.
 * @param dispatch Dispatches data to your app.
 * @param exit Called when all matching vacancies have left the DOM.
 */
export interface PatternCallback<A> {
    (matchObject: PatternMatch, dispatch: Dispatcher<A>, exit: ExitPromise): void | Promise<void>;
}
/**
 * Callback that executes when a vacancy matches a regex pattern.
 *
 * @typeparam A The type of object that you'll dispatch to your app.
 * @param matchArray The result of matching the vacancy with the regex.
 * @param dispatch Dispatches data to your app.
 * @param exit Called when all matching vacancies have left the DOM.
 */
export interface RegExpCallback<A> {
    (matchArray: RegExpMatchArray, dispatch: Dispatcher<A>, exit: ExitPromise): void | Promise<void>;
}
/**
 * Callback that executes when any vacancy is found.
 *
 * @typeparam A The type of object that you'll dispatch to your app.
 * @param vacancy The vacancy that was found.
 * @param dispatch Dispatches data to your app.
 * @param exit Called when all matching vacancies have left the DOM.
 */
export interface WildcardCallback<A> {
    (vacancy: string, dispatch: Dispatcher<A>, exit: ExitPromise): void | Promise<void>;
}
/**
 * Used to dispatch an object of type `<A>` to any subscribers.
 *
 * @typeparam A The type of object to be dispatched.
 */
export interface Dispatcher<A> {
    (action: A): void;
}
/**
 * This is your main API entry point to this library.
 *
 * An instance of this class manages your vacancy observers,
 * including methods for adding observers and listening on a
 * DOM element. Typically the instance would be created at
 * application startup time and would last for the duration
 * of the app.
 *
 * Internally, it creates a `MutationObserver` which does
 * the actual work of detecting vacancies.
 *
 * For convenience, methods on this class are chainable,
 * allowing patterns like this:
 *
 * ```js
 * const vacancies = Motel.create()
 *   .observe(...observer callback...)
 *   .observe(...observer callback...)
 *   .observe(...observer callback...)
 *   .subscribe(action => store.dipatch(action))
 *   .connect(document.querySelector('#root'));
 * ```
 *
 * ...however note that it may be better to break these
 * calls across different modules.
 *
 * @typeparam A The output type of this Motel instance.
 *   That is, the type of object you'll dispatch from your
 *   observer callbacks. For example, in a Redux app this
 *   would probably be your Redux action type.
 */
export default class Motel<A = any> {
    /**
     * Create a new instance with the given options.
     *
     * @typeparam A The type of object that you'll dispatch to
     *   your app from your vacancy observers.
     */
    static create<AA = any>(opts?: MotelOptions): Motel<AA>;
    private readonly telemetry;
    private readonly send;
    private readonly observers;
    private readonly subscriptions;
    private lifecycle?;
    private constructor();
    /**
     * Create a wildcard observer that sees every vacancy.
     *
     * @param wildcard The special `"*"` string, which represents
     *   a pattern that matches everything.
     */
    observe(wildcard: '*', handler: WildcardCallback<A>): Motel<A>;
    /**
     * Create an observer that matches vacancies based on the
     * given string pattern. Pattern matching follows the rules of
     * a [UrlPattern](https://www.npmjs.com/package/url-pattern)
     * object.
     *
     * @param stringPattern A pattern string as described
     *   [here](https://www.npmjs.com/package/url-pattern).
     */
    observe(stringPattern: string, handler: PatternCallback<A>): Motel<A>;
    /**
     * Create an observer that matches vacancies based on a regex.
     *
     * @param regex A pattern string as described
     *   [here](https://www.npmjs.com/package/url-pattern).
     */
    observe(regex: RegExp, handler: RegExpCallback<A>): Motel<A>;
    /**
     * Connect a `MutationObserver` to the given root DOM element and begin watchng for vacancies.
     * It also synchronously finds pre-existing vacancies in the root's subtree and reports them to observers.
     * Observers should thus be attached before calling this method, or some vacancies may go undetected.
     * This is especially relevant if you're using SSR.
     *
     * @param root Vacancies occurring on or anywhere below
     *   this element will be observed.
     */
    connect(root: Element): Motel<A>;
    /**
     * Stop listening for vacancies.
     */
    disconnect(): Motel<A>;
    /**
     * Subscribe to the output of this instance. That is, a stream
     * of objects of type `<A>`. Every dispatched object will be
     * seen by the subscriber.
     *
     * @param subscriber A callback function which receives
     *   objects of type `A` dispatched from your vacancy
     *   observers.
     */
    subscribe(subscriber: Dispatcher<A>): Motel<A>;
    /** @hidden */
    _publish(vacancy: string, exitProm: Promise<void>): void;
}
//# sourceMappingURL=index.d.ts.map