"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.VACANCY_ATTRIBUTE = void 0;
const url_pattern_1 = __importDefault(require("url-pattern"));
const assert_never_1 = __importDefault(require("./assert-never"));
const element_lifecycle_1 = require("./element-lifecycle");
const gate_keeper_1 = require("./gate-keeper");
const telemetry_1 = __importStar(require("./telemetry"));
/** The name of the vacancy attribute. */
exports.VACANCY_ATTRIBUTE = 'data-vacancy';
const DEFAULT_OPTIONS = {
    telemetryLevel: 'warn',
    telemetryHandler: telemetry_1.consoleTelemetryHandler,
};
/**
 * This is your main API entry point to this library.
 *
 * An instance of this class manages your vacancy observers,
 * including methods for adding observers and listening on a
 * DOM element. Typically the instance would be created at
 * application startup time and would last for the duration
 * of the app.
 *
 * Internally, it creates a `MutationObserver` which does
 * the actual work of detecting vacancies.
 *
 * For convenience, methods on this class are chainable,
 * allowing patterns like this:
 *
 * ```js
 * const vacancies = Motel.create()
 *   .observe(...observer callback...)
 *   .observe(...observer callback...)
 *   .observe(...observer callback...)
 *   .subscribe(action => store.dipatch(action))
 *   .connect(document.querySelector('#root'));
 * ```
 *
 * ...however note that it may be better to break these
 * calls across different modules.
 *
 * @typeparam A The output type of this Motel instance.
 *   That is, the type of object you'll dispatch from your
 *   observer callbacks. For example, in a Redux app this
 *   would probably be your Redux action type.
 */
class Motel {
    constructor(opts) {
        const allOpts = Object.assign(Object.assign({}, DEFAULT_OPTIONS), opts);
        this.telemetry = new telemetry_1.default(allOpts.telemetryLevel, allOpts.telemetryHandler);
        const observers = [];
        const subscriptions = [];
        const send = createPublishFunc(subscriptions, this.telemetry);
        this.send = send;
        this.observers = observers;
        this.subscriptions = subscriptions;
        this.telemetry.send('debug', 'instance created');
    }
    /**
     * Create a new instance with the given options.
     *
     * @typeparam A The type of object that you'll dispatch to
     *   your app from your vacancy observers.
     */
    static create(opts = DEFAULT_OPTIONS) {
        return new Motel(opts);
    }
    /**
     * @param handler A callback function that runs whenever
     *   a vacancy is found.
     */
    observe(matcher, handler) {
        const { observers } = this;
        if (matcher === '*') {
            this.telemetry.send('debug', 'adding wildcard observer');
            observers.push({ is: 'wildcard', handler });
        }
        else if (typeof matcher === 'string') {
            const pattern = new url_pattern_1.default(matcher);
            this.telemetry.send('debug', `adding string pattern observer: ${matcher}`);
            observers.push({ is: 'pattern', pattern, handler });
        }
        else {
            const regex = matcher;
            this.telemetry.send('debug', `adding regex observer: ${regex}`);
            observers.push({ is: 'regex', regex, handler });
        }
        return this;
    }
    /**
     * Connect a `MutationObserver` to the given root DOM element and begin watchng for vacancies.
     * It also synchronously finds pre-existing vacancies in the root's subtree and reports them to observers.
     * Observers should thus be attached before calling this method, or some vacancies may go undetected.
     * This is especially relevant if you're using SSR.
     *
     * @param root Vacancies occurring on or anywhere below
     *   this element will be observed.
     */
    connect(root) {
        if (this.lifecycle) {
            this.telemetry.send('error', 'instance is already connected');
            return this;
        }
        const id = root.id ? `#${root.id}` : '';
        const className = root.className ? `.${root.className}` : '';
        const nodeDescriptor = `${root.nodeName.toLowerCase()}${id}${className}`;
        this.telemetry.send('debug', `connecting instance to DOM node: ${nodeDescriptor}`);
        const gateKeeper = new gate_keeper_1.GateKeeper();
        this.lifecycle = element_lifecycle_1.ElementLifecycle.of(root, exports.VACANCY_ATTRIBUTE)
            .on('enter', (el, vacancy) => __awaiter(this, void 0, void 0, function* () {
            this.telemetry.send('silly', `incrementing vacancy: ${vacancy}`);
            const exitProm = gateKeeper.incr(vacancy);
            if (exitProm) {
                // exits proceed async
                // so entrances should too
                yield tick();
                this._publish(vacancy, exitProm);
            }
        }))
            .on('exit', (el, vacancy) => {
            this.telemetry.send('silly', `decrementing vacancy: ${vacancy}`);
            gateKeeper.decr(vacancy);
        })
            .start();
        return this;
    }
    /**
     * Stop listening for vacancies.
     */
    disconnect() {
        if (!this.lifecycle) {
            this.telemetry.send('error', 'instance is not connected');
            return this;
        }
        this.telemetry.send('debug', 'disconnecting instance');
        this.lifecycle.stop();
        delete this.lifecycle;
        return this;
    }
    /**
     * Subscribe to the output of this instance. That is, a stream
     * of objects of type `<A>`. Every dispatched object will be
     * seen by the subscriber.
     *
     * @param subscriber A callback function which receives
     *   objects of type `A` dispatched from your vacancy
     *   observers.
     */
    subscribe(subscriber) {
        this.telemetry.send('debug', 'adding subscriber');
        const { subscriptions } = this;
        subscriptions.push(subscriber);
        return this;
    }
    /** @hidden */
    _publish(vacancy, exitProm) {
        this.telemetry.send('debug', `entering vacancy: ${vacancy}`);
        exitProm.then(() => this.telemetry.send('debug', `exiting vacancy: ${vacancy}`));
        const { observers, send } = this;
        const proms = [];
        for (let observer of observers) {
            switch (observer.is) {
                case 'wildcard': {
                    const { handler } = observer;
                    this.telemetry.send('debug', `handling wildcard vacancy: ${vacancy}`);
                    try {
                        proms.push(handler(vacancy, send, exitProm));
                    }
                    catch (ex) {
                        proms.push(Promise.reject(ex));
                    }
                    break;
                }
                case 'pattern': {
                    const { pattern, handler } = observer;
                    const match = processMatch(pattern.match(vacancy));
                    if (match) {
                        this.telemetry.send('debug', `handling string pattern vacancy: ${vacancy}`, match);
                        try {
                            proms.push(handler(match, send, exitProm));
                        }
                        catch (ex) {
                            proms.push(Promise.reject(ex));
                        }
                    }
                    break;
                }
                case 'regex': {
                    const { regex, handler } = observer;
                    const match = vacancy.match(regex);
                    if (match) {
                        this.telemetry.send('debug', `handling regex vacancy: ${vacancy}`, match);
                        try {
                            proms.push(handler(match, send, exitProm));
                        }
                        catch (ex) {
                            proms.push(Promise.reject(ex));
                        }
                    }
                    break;
                }
                default: {
                    this.telemetry.send('critical', 'untyped observer', observer);
                    assert_never_1.default(observer);
                }
            }
        }
        if (proms.length === 0) {
            this.telemetry.send('warn', `unobserved vacancy: ${vacancy}`);
        }
        Promise.all(proms).catch(err => this.telemetry.send('error', `error while handling vacancy: ${vacancy}`, err));
    }
}
exports.default = Motel;
function createPublishFunc(subscriptions, telemetry) {
    return (action) => {
        for (let sub of subscriptions) {
            try {
                telemetry.send('debug', 'publishing action', action);
                sub(action);
            }
            catch (ex) {
                telemetry.send('error', 'error while publishing action', ex);
            }
        }
    };
}
function processMatch(match) {
    if (match) {
        const result = {};
        for (let [key, val] of Object.entries(match)) {
            if (Array.isArray(val)) {
                val = val[0];
            }
            if (typeof val === 'string') {
                result[key] = val;
            }
            else {
                result[key] = `${val}`;
            }
        }
        return result;
    }
    else {
        return null;
    }
}
function tick() {
    return new Promise(r => setImmediate(r));
}
